/**
Performance test for ARB_fragment_program OpenGL extension.
Generates a zillion small ARB_fragment_program strings, to 
test out:
	- Performance of various instructions
	- Limits on instruction counts
The performance measurement business is quite tricky, and 
drivers are beginning to heavily optimize out computations--
treat results with caution!

Orion Sky Lawlor, olawlor@acm.org, 2005/12/02 (Public Domain)
*/
/* This file is #included by a larger performance testing framework.
  Requires OpenGL, GLUT, and GLEW (GL Extensions Wrangler).
  The only custom dependency here is "getTime" (returns wall-clock 
  in seconds) and "oglPrint" (debugging printout). 
#include <string>

/* Namespace to hide all the fragment instruction measurement routines */
namespace fragInst_namespace {

static void makeVertex(double x,double y) {
        glTexCoord2f(x*0.05,-y*0.05);
        glVertex3f(x,y,0);
}

/** Assumes screen is set up so 1.0 coordinate equals one pixel. */
static void drawPoly(double size) {
	glEnable(GL_FRAGMENT_PROGRAM_ARB);
	glEnable(GL_TEXTURE_2D);
	glBegin(GL_QUADS);
	makeVertex(0,0); 
	makeVertex(0,size); 
	makeVertex(size,size);
	makeVertex(size,0); 
	glEnd();
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_FRAGMENT_PROGRAM_ARB);
}

struct shaderTimes {
	double upload; // time per upload
	double poly; // time per polygon
	double pixel; // time per pixel (seconds)
};

static shaderTimes timeShader(const std::string &shaderGuts,bool withTex,bool testOnly=false) 
{
	std::string prefix="";
	if (withTex) prefix="TEX t0, fragment.texcoord[0], texture[0], 2D;\n";
	std::string shader="!!ARBfp1.0\n"
		"TEMP r0,r1,r2, t0,t1,t2;\n"
		"PARAM c0 = {0.90, 0.91, 0.92, 0.93};\n"
		"PARAM c1 = {0.95, 0.96, 0.97, 0.98};\n"
		"PARAM c2 = {0.85, 0.86, 0.87, 0.88};\n"
		+prefix+shaderGuts+""
		"END\n";
	
	unsigned int shader_num;
	
	shaderTimes ret;
	double start;
	start=getTime();
	glGenProgramsARB(1, &shader_num);
	glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, shader_num);
	oglCheck("Prior to fragment shader upload");
	
	glProgramStringARB(GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, 
		shader.size(), shader.c_str());
	ret.upload=getTime()-start;
	
	if (testOnly) {
		if (glGetError()!=GL_NO_ERROR) { 
		/* an error during shader upload: normally caused by hitting a fixed resource limit */
			ret.pixel=-1;
		} else {
			ret.pixel=1;
		}
		return ret;
	}
	oglCheck(("fragment shader upload: "+shaderGuts).c_str()); // +shaderGuts.c_str()
	
	glClear(GL_COLOR_BUFFER_BIT+GL_DEPTH_BUFFER_BIT); glFlush();
	
	/* Compute time to draw a 1-pixel-across polygon */
	drawPoly(1);
	start=getTime();
	drawPoly(1); glFinish();
	ret.poly=getTime()-start;
	
	/* Compute time to draw a bunch of big quads 
		FIXME: determine "rep" dynamically, average out error, etc.
	*/
	double sz=500.0; int rep=10;
	start=getTime();
	for (int i=0;i<rep;i++) drawPoly(sz); 
	glFinish();
	ret.pixel=(getTime()-start)/(sz*sz*rep); /* time per pixel */
	
	oglPrint(osl::Vector3d(10,sz+10,0), shaderGuts.c_str());
	
	glutSwapBuffers(); 
	return ret;
}

/// Aspects of a shader: swizzles, negates, saturates, etc.
typedef enum {
	shaderFlag_none=0,
	shaderFlag_swz=(1<<1),
	shaderFlag_neg=(1<<2),
	shaderFlag_sat=(1<<3),
	shaderFlag_msk=(1<<4),
	shaderFlag_scalar=(1<<10) // scalar instruction
} shaderFlagEnum;
typedef unsigned int shaderFlags_t;

/**
  Describes one kind of fragment instruction.
*/
struct fragInst_info {
	/** Instruction mnemonic, like "MUL" or "ADD". */
	const char* name;
	
	/** Number of arguments, not counting output argument or suffix. */
	int nArg; 
	
	/** Junk to put after instruction during testing:
	   e.g., "" for most instructions, but ",texture[0],2D" for TEX
	*/
	const char* suffix; 
	
	/** Human-readable instruction description */
	const char* function;
	
	/** Flags: shader aspects *not* to add */
	shaderFlags_t flags;
};

/**
  All the different types of fragment instruction.
*/
static const fragInst_info fragInstructions[]={
	{"ABS",1,"","Absolute value (fabs(A))"},
	{"ADD",2,"","Addition (A+B)"},
	{"CMP",3,"","Compare (A<0?B:C)"},
	{"COS",1,"","Scalar cosine (cos(A))",shaderFlag_scalar},
	{"DP3",2,"","Dot product, 3 components"},
	{"DP4",2,"","Dot product, 4 components"},
	{"DPH",2,"","Dot product3 plus B's w"},
	{"DST",2,"","Funky distance computation"},
	{"EX2",1,"","Scalar base-2 exponentiation (pow(2,A))",shaderFlag_scalar},
	{"FLR",1,"","Floor (floor(A))"},
	{"FRC",1,"","Fraction (A-floor(A))"},
	{"LG2",1,"","Scalar base-2 logarithm (log(A)/log(2))",shaderFlag_scalar},
	{"LIT",1,"","Funky lighting calculation"},
	{"LRP",3,"","Linear interpolation (A*B+(1-A)*C)"},
	{"MAD",3,"","Multiply-add (A*B+C)"},
	{"MAX",2,"","Maximum (max(A,B))"},
	{"MIN",2,"","Minimum (min(A,B))"},
	{"MOV",1,"","Move (A)"},
	{"MUL",2,"","Multiply (A*B)"},
	{"POW",2,"","Scalar power function (pow(A,B))",shaderFlag_scalar},
	{"RCP",1,"","Scalar reciprocal estimate (1.0/A)",shaderFlag_scalar},
	{"RSQ",1,"","Scalar inverse-square-root estimate (1.0/sqrt(A))",shaderFlag_scalar},
	{"SCS",1,"","Scalar sin & cosine (sin(A),cos(A))",shaderFlag_scalar},
	{"SGE",2,"","Set if greater than or equal to (A>=B)",shaderFlag_sat},
	{"SLT",2,"","Set if less than (A<B)",shaderFlag_sat},
	{"SIN",1,"","Scalar sine (sin(A))",shaderFlag_scalar},
	{"SUB",2,"","Subtract (A-B)"},
	{"SWZ",1,",-y,-z,-w,-x","Swizzle",shaderFlag_swz|shaderFlag_neg},
	{"TEX",1,",texture[0],2D","Texture sample"},
	{"TXP",1,",texture[0],2D","Texture sample with perspective divide"},
	{"TXB",1,",texture[0],2D","Texture sample with LOD bias"},
	{"XPD",2,"","Vector cross product (A x B)"},
	{NULL}
};

// Flags to turn on during run:
shaderFlags_t shaderFlags=shaderFlag_none;
static const char *shaderFlagNames[]={
	"","Swizzle","Negate","Saturate","Writemask","??"
};

/**
  Return the arguments taken by this instruction, given that
  it will return returnTo.
*/
static std::string shaderArgs(const fragInst_info &inst,std::string returnTo,bool firstTime=false)
{
	std::string args=returnTo;
	if ((shaderFlags&~inst.flags)&shaderFlag_msk) args+=".xyz";
	for (int i=0;i<inst.nArg;i++) {
		args+=",";
		if ((shaderFlags&~inst.flags)&shaderFlag_neg) args+="-";
		if (i==0) { /* first argument is t0 or r0 */
			if(firstTime) args+="t0"; 
			else args+="r0";
		} else { /* subsequent arguments are all constants */
			char id[2];
			id[0]='0'+i; id[1]=0;
			args+="c"+std::string(id);
		}
		if (inst.flags&shaderFlag_scalar)
			args+=".r"; // take red component
		else if ((shaderFlags&~inst.flags)&shaderFlag_swz)
			args+=".yzwx";
	}
	return args;
}

/** 
  Return the time (in seconds) per pixel when executing this instruction k times. 
*/
static double timePer(const fragInst_info &inst,int n,bool testOnly=false) 
{
	std::string shaderGuts="";
	for (int i=0;i<n;i++)
	{
		shaderGuts+=inst.name;
		if ((shaderFlags&~inst.flags)&shaderFlag_sat)
			shaderGuts+="_SAT";
		const char *ret=" result.color";
		if (i<n-1) ret=" r0";
		shaderGuts+=shaderArgs(inst,ret,i==0)+inst.suffix+";\n";
	}
	shaderTimes t=timeShader(shaderGuts,true,testOnly);
	return t.pixel;
}

/// Return true if these two timings, in nanoseconds, 
///   are "about the same" and hence not worth printing out.
bool small_difference(double a,double b) {
	double diff=fabs(a-b);
	if (diff<0.05) return true; // difference is small in absolute terms
	if (diff/a<0.1) return true; // difference is small in relative terms
	return false; // not a small difference 
}

/*
  Loop over the fragment instructions, and time each one.
*/
void timeFragInst(void) 
{
	for (int inst_i=0; fragInstructions[inst_i].name!=0; inst_i++) /* instruction types */
	{
		const fragInst_info &inst=fragInstructions[inst_i];
		double t_1_base=0, t_mea_base=0; int max_base=0;
		int nprint=0;
		printf("  %s:\t",inst.name);
	for (int sc=0;sc<=4;sc++) /* flag types: none, saturate, swizzle, etc */
	{
		shaderFlags=(shaderFlags_t)(1<<sc);
		const char *flagName=shaderFlagNames[sc];
		
		/// Figure out how many times we can execute this instruction.
		///   Uses an exponential search.
		int max=0, next; /* max works, but next may not */
		while (1) {
			next=(int)(max*2);
			if (next<=max) next=max+1;
			if (timePer(inst,next,true)>=0) max=next; /* worked */
			else break; /* didn't work-- can't compile "next" */
			if (max>=1024) break; /* over 1000 is plenty */
		}
		/// Refine number-of-executions search for low numbers:
		if (max<=32) {
			int end=max*2;
			for (next=max;next<end;next++) {
				if (timePer(inst,next,true)>=0) max=next; /* worked */
				else break; /* didn't work-- can't compile "next" */
			}
		}
		
		/// Measure the speed of this instruction.
		int mea=max;
		if (mea>20) mea=20;
		
		double t_1=timePer(inst,1)*1.0e9;
		double t_mea=(timePer(inst,mea)*1.0e9-t_1)/(mea-1);
		if (sc==0) { /* base timings */
			t_1_base=t_1;
			t_mea_base=t_mea;
			max_base=max;
		} else { /* non-base mode: only show if it's interesting */
			if (small_difference(t_mea,t_mea_base))
				continue; /* same as base-- don't print */
		}
		
		printf("%5.2f ns %s\t",t_mea,flagName);
		if (sc==0) printf("%4d times\t",max_base);
		nprint++;
	}
		printf("\n");
		fflush(stdout);
	}
}
};


/**
  Time all the different fragment instructions.
*/
void timeFragmentInstructions(void) {
	if (!GLEW_ARB_fragment_program) { 
		printf("ARB_fragment_program extension not available.\n");
		return;
	}
	printf("\n=== ARB_fragment_program instructions ===\n");
	using namespace fragInst_namespace;
	
	oglTexture tex(image,imageWidth,imageHeight, oglTexture_mipmap,
		GL_RGBA, GL_UNSIGNED_BYTE);
	tex.bind();
	
	timeFragInst();
}

